/*
 * Relays complete environment and stdin from CGI request plus a fixed
 * password to a daemon. Relays the reply from a daemon back to the web
 * server.
 *
 * (c) September 2000 Jol Savignon <js@neosystem.com>
 * from cgi-proxy.c and cgi-proxy2.c
 * written by Samuli Krkkinen <skarkkai@woods.iki.fi> at July 2000.
 *
 * Released under GPL.
 *
 */

#include <stdio.h>
#if defined (__sun) || defined (_AIX)
  #include <strings.h>
#else
  #include <string.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>

#define CONF_FILE "/etc/wcal.conf"
#define ADDRESS "127.0.0.1"
#define PORT "13134"

#define SOCKET_WRITE_BUF 4096
#define SOCKET_READ_BUF 1024
#define BUFFER_INITIAL_SIZE 16*1024

#define min(x,y) ((x) < (y) ? (x) : (y))
#define max(x,y) ((x) > (y) ? (x) : (y))

char *password;
char *address;
char *port;

void print_http_header ()
{
  printf ("Content-Type: text/html\n\n");
  return;
}

void print_html_error_pre_stuff ()
{
  printf ("<HTML>\n<HEAD><TITLE>Internal Error</TITLE></HEAD>\n\n");
  printf ("<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\" LINK=\"#0000b0\" VLINK=\"#0000b0\" ALINK=\"#0000b0\">\n");
  return;
}

void print_html_error_post_stuff ()
{
  printf ("</BODY>\n</HTML>\n");
  return;
}

/*
 * Print a HTTP/HTML error message to stdout
 */
void print_error (char *fmt, ...)
{
  va_list	ap;		/* points to each unnamed arg in turn */

  if (! fmt)
    return;
  va_start (ap, fmt);
  print_http_header ();
  print_html_error_pre_stuff ();
  printf ("<H2>Internal error: ");
  vprintf (fmt, ap);
  printf ("</H2>\n");
  print_html_error_post_stuff ();

  return;
}

/*
 * Print text message to a socket
 */

void socket_print_error (int s, void *data)
{
  write (s, data, strlen (data));
}

char *remove_leading_trailing_space (char *s)
{
  char *start;
  while (*s == ' ')
    s++;
  start = s;
  if (! *s)
    return start;
  while (*s)
    s++;
  s--;
  while (*s == ' ')
    s--;
  *(s+1) = '\000';
  return start;
}

static struct b_or_pb {
  char	*data;
  int		size;
  int		used;
  int		tail;
} b, pb;

void * my_malloc (size_t size)
{
  void *p;
	
  p = malloc (size);
  if (! p) {
    socket_print_error (1, "malloc: ");
    socket_print_error (1, strerror(errno));
    exit (1);
  }
  return (p);
}

/*
 * Add data to lilo
 */
void lilo_add (void *data, int size)
{
  /*
   * If the lilo is empty, this is the first time we're called
   */
  if (b.size == 0) {
    b.data = my_malloc (BUFFER_INITIAL_SIZE);
    b.size = BUFFER_INITIAL_SIZE;
    b.used = 0;
    b.tail = 0;
  }

  /*
   * If we're running out of space, double the lilo
   */
  if (b.used + size > b.size) {
    void *newbuf;
    newbuf = my_malloc (b.size * 2);
    memcpy (newbuf, b.data + b.tail, b.used - b.tail);
    b.data = newbuf;
    b.size = b.size * 2;
    b.used = b.used - b.tail;
    b.tail = 0;
  }
	
  /*
   * Now we are guaranteed to have enough buffer space.
   * Copy the data into buffer.
   */
  memcpy (b.data + b.used, data, size);
  b.used += size;
}

/*
 * Return a pointer to some amount of data
 */
void lilo_get (void **data, int *size)
{
  *data = b.data + b.tail;
  *size = min (b.used - b.tail, SOCKET_WRITE_BUF);
}

/*
 * Remove some data from lilo by advancing tail
 */
void lilo_remove (int size)
{
  b.tail += size;
  if (b.tail > b.used)
    b.tail = b.used;
}

/*
 * Add data into lilo, then try write data from lilo to
 * the socket, and return immediately if the write would block.
 */
void write_nonblocking (void *data, int size, int s)
{
  int wrote;
	
  /*
   * Add new data into head of the infinite buffer
   */
  lilo_add (data, size);

  /*
   * Get some chunk of data from the tail of the buffer
   */
  lilo_get (&data, &size);
	
  /*
   * If buffer gave us zero bytes, return
   */
  if (! size)
    return;
	
  /*
   * Otherwise try write it into the nonblocking output socket
   */
  wrote = write (s, data, size);
  if (wrote == -1) {

    /*
     * If there is no space in buffer, we get EAGAIN
     */
    if (errno == EAGAIN)
      return;

    /*
     * Else a real error has occured, bail out
     */
    socket_print_error (1, "write(s3): ");
    socket_print_error (1, strerror(errno));
    exit (1);
  }
	
  /*
   * We got stuff written, remove that many bytes from buffer tail
   */
  lilo_remove (size);

  return;
}

void my_write (int fd, const void *buf, size_t len)
{
  size_t wrote;
  wrote = write (fd, buf, len);
  if (wrote != len) {
    if (wrote == -1) {
      print_error ("write(%d): %s", fd, strerror(errno), fd);
    } else {
      print_error ("write(%d) failed to write entire buffer", fd);
    }
    exit (1);
  }
}

void post_buffer_write_all (int fd) {
  int left;
	
  while ((left = min (SOCKET_WRITE_BUF, pb.used - pb.tail)) > 0) {
    my_write (fd, pb.data + pb.tail, left);
    pb.tail += left;
  }
}

void add_post_buffer (void *buffer, int size)
{
  /*
   * If the buffer is empty, this is the first time we're called
   */
  if (pb.size == 0) {
    pb.data = my_malloc (BUFFER_INITIAL_SIZE);
    pb.size = BUFFER_INITIAL_SIZE;
    pb.used = 0;
    pb.tail = 0;
  }

  /*
   * If we're running out of space, double the buffer
   */
  if (pb.used + size > pb.size) {
    void *newbuf;
    newbuf = my_malloc (pb.size * 2);
    memcpy (newbuf, pb.data, pb.used);
    pb.data = newbuf;
    pb.size = pb.size * 2;
  }

  /*
   * Now we are guaranteed to have enough buffer space.
   * Copy the data into buffer.
   */
  memcpy (pb.data + pb.used, buffer, size);
  pb.used += size;
}


/*
 * Write all of the lilo to the socket in blocking mode.
 */
void write_blocking (int s)
{
  int wrote;
	
  while (b.tail < b.used) {
    wrote = write (s, b.data + b.tail, min (b.used - b.tail, SOCKET_WRITE_BUF));
    if (wrote == -1) {
      socket_print_error (1, "write(s4): ");
      socket_print_error (1, strerror(errno));
      exit (1);
    }
    b.tail += wrote;
  }
}

int get_conf_line (FILE *file, char **key, char **value)
{
  static char buffer[10240];
  char *lkey, *lvalue;
  int i;
	
  if (! fgets (buffer, 10240, file)) {
    return (0);
  }
  for (i = 0; buffer[i] != '\000'; i++) {
    if (buffer[i] == '\t' || buffer[i] == '\n') {
      buffer[i] = ' ';
    }
  }
  lkey = buffer;
  lvalue = strchr (buffer, '=');
  if (lvalue == NULL) {
    lvalue = "";
  } else {
    *lvalue = '\000';
    lvalue++;
  }
  lkey = remove_leading_trailing_space (lkey);
  for (i = 0; lkey[i] != '\000'; i++) {
    lkey[i] = tolower (lkey[i]);
  }
  lvalue = remove_leading_trailing_space (lvalue);
  *key = lkey;
  *value = lvalue;
  return (1);
}

void read_config ()
{
  FILE *file;
  char *key, *value;
	
  file = fopen (CONF_FILE, "r");
  if (! file) {
    print_error ("can't open " CONF_FILE " for reading");
    exit (1);
  }
  address = NULL;
  port = NULL;
  password = NULL;
  while (get_conf_line (file, &key, &value)) {
    if (strcmp (key, "password") == 0) {
      password = strdup (value);
    } else if (strcmp (key, "address") == 0) {
      address = strdup (value);	/* host name/address on which wcald runs */
    } else if (strcmp (key, "port") == 0) {
      port = strdup (value);		/* port on which wcald listens */
    }
  }
  if (! password) {
    print_error ("can't find password from configuration file");
    exit (1);
  }
  if (! address) {
    address = ADDRESS;	/* default value */
  }
  if (! port) {
    port = PORT;		/* default value */
  }
  fclose (file);

  return;
}

int main (int argc, char **argv, char **envp)
{
  int s;
  ssize_t got;
  struct sockaddr_in sin;
  char read_buf[SOCKET_READ_BUF];
  char curwd[1024];
  char *basename;
  char buf[8192];

  read_config ();

  /*
   * Read POST data into a buffer
   */
	
  while ((got = read (0, read_buf, SOCKET_READ_BUF)) != 0) {
    if (got == -1) {
      print_error ("read(-1): %s", strerror(errno));
      exit (1);
    }
    add_post_buffer (read_buf, got);
  }

  /*
   * Create the socket
   */
  s = socket (AF_INET, SOCK_STREAM, 0);
  if (s == -1) {
    print_error ("socket: %s", strerror(errno));
    exit (1);
  }

  /*
   * Connect
   */
  bzero (&sin, sizeof (sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = inet_addr (address);
  sin.sin_port = htons ((u_short) atoi (port));
  if (connect (s, (struct sockaddr *) &sin, sizeof (sin)) < 0) {
    print_error ("connect: %s<P>The server process probably isn't running at host %s, port %s.\n", strerror(errno), address, port);
    exit (1);
  }

  /*
   * Write password for authentication
   */
  if (! getcwd (curwd, (size_t) 1024)) {
    perror ("getcwd");
    exit (1);
  }
  basename = strrchr (curwd, '/');
  basename++;

  my_write (s, "__password=", strlen ("__password="));
  my_write (s, password, strlen (password)); my_write (s, "\n", 1);

  my_write (s, "__user=", strlen ("__user="));
  my_write (s, basename, strlen (basename)); my_write (s, "\n", 1);

  sprintf (buf, "%d", (int)getegid());
  my_write (s, "__gid=", strlen ("__gid="));
  my_write (s, buf, strlen (buf)); my_write (s, "\n", 1);

  /*
   * Write environment
   */
  for (; *envp != NULL; envp++) {
    static char buf[256*1024];
    int data_len;
		
    data_len = strlen (*envp);
    if (data_len > 255*1024) {
      print_error ("too long environment variable");
      exit (1);
    }
    if (strchr (*envp, '\n'))
      continue;	/* don't send variables with newlines in them, such as Apache 1.3+ Serversignature */
    memcpy (buf, *envp, data_len);
    buf[data_len] = '\n';
    data_len++;
    buf[data_len] = '\0';
    my_write (s, buf, data_len);
  }

  /*
   * Write POST data that was read from STDIN
   */
  post_buffer_write_all (s);
  
  if (shutdown(s, 1) == -1) {;		/* close the writing side */
  print_error ("shutdown(%d, 1): %s", s, strerror(errno));
  exit (1);
  }

  /*
   * Flush stdout, and start writing to corresponding filedescriptor 1
   * only from now on.
   */
  if (fflush (stdout) != 0) {
    print_error ("fflush(stdout): %s", strerror(errno));
    exit (1);
  }
  
  /*
   * Move the data from daemon to web server in following way
   *  - read one buffer in block way
   *    - if got data
   *      - try writing it in non-blocking way
   *         - if fails, buffer the data
   *    - if EOF
   *      - write the rest of the buffer in blocking way
   *  - start over
   */
  if (fcntl (1, F_SETFL, O_NONBLOCK) == -1) {
    socket_print_error (1, "fcntl: ");
    socket_print_error (1, strerror(errno));
    exit (1);
  }
  
  while ((got = read (s, read_buf, SOCKET_READ_BUF)) != 0) {
    if (got == -1) {
      socket_print_error (1, "read(s): ");
      socket_print_error (1, strerror(errno));
      exit (1);
    }
    write_nonblocking (read_buf, got, 1);
  }
  
  /*
   * Write the rest of the data in blocking way
   */
  if (fcntl (1, F_SETFL, 0) == -1) {
    socket_print_error (1, "fcntl2: ");
    socket_print_error (1, strerror(errno));
    exit (1);
  }
  write_blocking (1);
  
  close (s);
  return 0;
}


